/*----------------------------------------------------------------------*/
/* FatFs sample project for Atmel AVR (adapted for Propeller 2 by RJA)                (C)ChaN, 2018      */
/*----------------------------------------------------------------------*/


//#include <string.h>
#include "ff.h"
#include "diskio.h"
#include "rtc.h"
#include "sound.h"


#include <stdio.h>

#include <propeller2.h>
#include <stdlib.h>

// to get the frequency we desire, we
// specify P2_TARGET_MHZ then include "sys/p2es_clock.h"
// e.g. to get 160_000_000, define P2_TARGET_MHZ to 160
// p2es_clock.h defines constants _SETFREQ and _CLOCKFREQ
// which we can then use to set the clock
//
// note that this is not the only way to set the clock; you can
// certainly calculate the desired clock mode based on the frequency
// and pass the resulting mode and frequency directly to clkset()
// p2es_clock.h is just a convenience header
#define P2_TARGET_MHZ 300
#include "sys/p2es_clock.h"



BYTE Buff[4096];	/* Working buffer */

FATFS FatFs[2];		/* Filesystem object for each logical drive */
FIL File[2];		/* File object */
FFDIR Dir;			/* Directory object */  //RJA:  Fastspin renames DIR to FFDIR

FILINFO Finfo;
DWORD AccSize;				/* Work register for fs command */
WORD AccFiles, AccDirs;

BYTE RtcOk;				/* RTC is available */
volatile UINT Timer;	/* Performance timer (100Hz increment) */




//RJA fixing some things...
#define PSTR  //RJA replacing this with nothing...
#define xputc putchar
#define xgetc getchar
//#define xatoi atoi //Actually need to define a function for this...
#define xprintf printf
#define xputs puts

#define get_line gets
#define xsprintf sprintf


char xatoi(char** str, long* ret)
{//RJA: returns value of next parameter in string and then points to next parameter in string
	while ((**str == ' ') && (**str != 0))  //need to advance pointer past spaces?
	{
		(*str)++;
	}
	if (**str == 0)
		return 0; //failed, nothing there
	*ret = atoi(*str);
	while ((**str != ' ') && (**str != 0))  //need to advance pointer?
	{
		(*str)++;
	}
	return 1;
}






/*--------------------------------------------------------------------*/
/* User Provided Timer Function for FatFs module                      */
/*--------------------------------------------------------------------*/
/* This is a real time clock service to be called from FatFs module.  */
/* Any valid time must be returned even if the system does not        */
/* support a real time clock. This is not required when FatFs is      */
/* configured for FF_FS_READONLY or FF_FS_NORTC = 1.                  */
/*--------------------------------------------------------------------*/


DWORD get_fattime(void)  //Note:  RJA set FF_FS_NORTC to 0 in ffconf.h to enable RTC support
{
	RTC rtc;


	if (!RtcOk) return 0;

	/* Get local time */
	rtc_gettime(&rtc);

	/* Pack date and time into a DWORD variable */
	return	  ((DWORD)(rtc.year - 1980) << 25)
		| ((DWORD)rtc.month << 21)
		| ((DWORD)rtc.mday << 16)
		| ((DWORD)rtc.hour << 11)
		| ((DWORD)rtc.min << 5)
		| ((DWORD)rtc.sec >> 1);
}


/*--------------------------------------------------------------------------*/
/* Monitor                                                                  */

static
void put_dump (const BYTE *buff, DWORD ofs, BYTE cnt)
{
	BYTE i;


	printf("%08lX:", ofs);

	for(i = 0; i < cnt; i++)
		printf(" %02X", buff[i]);

	xputc(' ');
	for(i = 0; i < cnt; i++)
		xputc((buff[i] >= ' ' && buff[i] <= '~') ? buff[i] : '.');

	xputc('\n');
}


static
FRESULT scan_files (
	char* path		/* Pointer to the working buffer with start path */
)
{
	FFDIR dirs;
	FRESULT fr;
	int i;

	fr = f_opendir(&dirs, path);
	if (fr == FR_OK) {
		while (((fr = f_readdir(&dirs, &Finfo)) == FR_OK) && Finfo.fname[0]) {
			if (Finfo.fattrib & AM_DIR) {
				AccDirs++;
				i = strlen(path);
				path[i] = '/'; strcpy(path+i+1, Finfo.fname);
				fr = scan_files(path);
				path[i] = 0;
				if (fr != FR_OK) break;
			} else {
//				xprintf(PSTR("%s/%s\n"), path, Finfo.fname);
				AccFiles++;
				AccSize += Finfo.fsize;
			}
		}
	}

	return fr;
}



static
void put_rc(FRESULT rc)
{//RJA had to rewrite this due to fastspin compiler error
	const char* p;
	switch (rc)
	{
	case 0:
		p = "OK"; break;
	case 1:
		p = "DISK_ERR"; break;
		break;
	case 2:
		p = "INT_ERR"; break;
	case 3:
		p = "NOT_READY"; break;
	case 4:
		p = "NO_FILE"; break;
	case 5:
		p = "NO_PATH"; break;
	case 6:
		p = "INVALID_NAME"; break;
	case 7:
		p = "DENIED"; break;
	case 8:
		p = "EXIST"; break;
	case 9:
		p = "INVALID_OBJECT"; break;
	case 10:
		p = "WRITE_PROTECTED"; break;
	case 11:
		p = "INVALID_DRIVE"; break;
	case 12:
		p = "NOT_ENABLED"; break;
	case 13:
		p = "NO_FILE_SYSTEM"; break;
	case 14:
		p = "MKFS_ABORTED"; break;
	case 15:
		p = "TIMEOUT"; break;
	case 16:
		p = "LOCKED"; break;
	case 17:
		p = "NOT_ENOUGH_CORE"; break;
	case 18:
		p = "TOO_MANY_OPEN_FILES"; break;
	default:
		p = "Unknown Error."; 
			
	}
	printf(PSTR("rc=%u FR_%S\n"), rc, p);
}







/*-----------------------------------------------------------------------*/
/* Main                                                                  */
int main(void)
{
	//SimpleTests();  //So some simple things like write to a file, show directories, read from a file
	InteractiveMode();  //The standard interactive menu of things to do that comes with several FatFs examples

}

void ListDir(char* ptr)
{
	FFDIR Dir1;
	FRESULT fr;
	int p1, s1, s2;
	FATFS* fs;

	fr = f_opendir(&Dir1, ptr);
	printf("Dir %s open with result %d\n", ptr, fr);
	put_rc(fr);
	if (fr) {
		return;
	}
	p1 = s1 = s2 = 0;
	for (;;) {
		fr = f_readdir(&Dir1, &Finfo);
		if ((fr != FR_OK) || !Finfo.fname[0])
		{
			//printf("bad dir.\n");  //RJA you get here after the last directory entry...
			return;
		}
		if (Finfo.fattrib & AM_DIR) {
			s2++;
		}
		else {
			s1++; p1 += Finfo.fsize;
		}
		printf("%c%c%c%c%c %u/%02u/%02u %02u:%02u %9lu  %s\n",
			(Finfo.fattrib & AM_DIR) ? 'D' : '-',
			(Finfo.fattrib & AM_RDO) ? 'R' : '-',
			(Finfo.fattrib & AM_HID) ? 'H' : '-',
			(Finfo.fattrib & AM_SYS) ? 'S' : '-',
			(Finfo.fattrib & AM_ARC) ? 'A' : '-',
			(Finfo.fdate >> 9) + 1980, (Finfo.fdate >> 5) & 15, Finfo.fdate & 31,
			(Finfo.ftime >> 11), (Finfo.ftime >> 5) & 63,
			(DWORD)Finfo.fsize,
			Finfo.fname);
	}
}

int SimpleTests(void)
{
	UINT bw;
	FRESULT fr;
	FIL Fil;			/* File object needed for each open file */


	printf("FatFs Simple Tests Program.\n");
	sleep(1);

	for (int i = 1; i < 10; i++)
	{
		printf("trying to mount...\n");
		fr = f_mount(&FatFs[0], "/", 1);		/* Give a work area to the default drive */
		printf("Result of mount= %d\n", fr);
		put_rc(fr);
		if (fr == 0)
			break;
		sleep(1);
	}
	if (fr != 0)
	{
		printf("Failed to mount...");
		return;
	}

	//Change directory
	fr = f_chdir("/");
	printf("*Change directory, result= %d\n", fr);
	put_rc(fr);

	//Test write to file
	fr = f_open(&Fil, "newfile.txt", FA_WRITE | FA_CREATE_ALWAYS);	/* Create a file */
	printf("Result of file open= %d\n", fr);
	put_rc(fr);
	int pr = f_puts("First string in my file.  What great text for a file!\n", &Fil);
	//Close file, don't forget this!
	f_close(&Fil);

	//Test read a file
	fr = f_open(&Fil, "newfile.txt", FA_READ);
	printf("File open result= %d\n", fr);
	put_rc(fr);
	char buff[256];
	UINT nread;
	//Read from file
	fr = f_read(&Fil, &buff, 255, &nread);
	printf("File read result= %d, %d bytes read.\nRead Characters:  ", fr, nread);
	put_rc(fr);
	for (int i = 0; i < (int) nread; i++)
	{
		printf("%c", buff[i]);
	}
	printf("\n");
	//Close file, don't forget this!
	f_close(&Fil);

	//Create a new directory
	fr =f_mkdir("NEWDIR");
	printf("Create Directory Result= %d\n", fr);
	put_rc(fr);


	//Test write to another file
	fr = f_open(&Fil, "NEWDIR/newfile2.txt", FA_WRITE | FA_CREATE_ALWAYS);	/* Create a file */
	printf("File open result= %d\n", fr);
	put_rc(fr);
	pr = f_puts("First string in my file, again\n", &Fil);
	//Close file, don't forget this!
	f_close(&Fil);

	//Test write to another file with long file name
	fr = f_open(&Fil, "NEWDIR/SuperLongFileName.txt", FA_WRITE | FA_CREATE_ALWAYS);	/* Create a file */
	printf("File open result= %d\n", fr);
	put_rc(fr);
	pr = f_puts("This file has a very long file name...\n", &Fil);
	//Close file, don't forget this!
	f_close(&Fil);


	//List the directories.
	ListDir((char *) "/");
	ListDir((char *) "NEWDIR");
	

	printf("All Done with Simple Tests!\n\n");
}

int InteractiveMode(void)
{

	char line[120];
	char* ptr, * ptr2;
	long p1, p2, p3;
	LBA_t lba;
	FRESULT fr;
	BYTE b1, * bp;
	UINT s1, s2, cnt, blen = sizeof Buff;
	DWORD ofs, sect = 0;
	static const char* fst[] = { "", "FAT12", "FAT16", "FAT32", "exFAT" };
	FATFS* fs;
	RTC rtc;

	printf("FatFs Interactive Test Program.\n");
	printf("Enter ? at the prompt for help.\n\n");
	
	
	if (rtc_init() && rtc_gettime(&rtc)) {
		RtcOk = 1;
		xprintf(PSTR("Current time: %u/%u/%u %02u:%02u:%02u\n"), rtc.year, rtc.month, rtc.mday, rtc.hour, rtc.min, rtc.sec);
	}
	else {
		xputs(PSTR("RTC is not available.\n"));
	}
	

	
	const char* test = "13\0";
	int val;
	char str[20];
	char* ts = str;
	char* t1;
	char* t2;

	for (;;) {
		xputc('>');
		ptr = line;

		gets(ptr);//RJA get_line(ptr, sizeof line);
		switch (*ptr++) {

		case 'T':
			while (*ptr == ' ') ptr++;

			/* Quick test space */

			break;

		case 'd':
			switch (*ptr++) {
			case 'd':	/* dd <pd#> [<sector>] - Dump secrtor */
				if (!xatoi(&ptr, &p1)) break;
				if (!xatoi(&ptr, &p2)) p2 = sect;
				b1 = disk_read((BYTE)p1, Buff, (DWORD)p2, 1);
				if (b1) { xprintf(PSTR("rc=%d\n"), b1); break; }
				sect = (DWORD)p2 + 1;
				xprintf(PSTR("Sector:%lu\n"), (DWORD)p2);
				for (bp = Buff, ofs = 0; ofs < 0x200; bp += 16, ofs += 16) {
					put_dump(bp, ofs, 16);
				}
				break;

			case 'i':	/* di <pd#> - Initialize disk */
				if (!xatoi(&ptr, &p1)) break;
				xprintf(PSTR("rc=%d\n"), disk_initialize((BYTE)p1));
				break;

			case 's':	/* ds <pd#> - Show disk status */
				if (!xatoi(&ptr, &p1)) break;
				if (disk_ioctl((BYTE)p1, GET_SECTOR_COUNT, &lba) == RES_OK) {
					xprintf(PSTR("Drive size: %lu sectors\n"), (DWORD)lba);
				}
				if (disk_ioctl((BYTE)p1, GET_BLOCK_SIZE, &p2) == RES_OK) {
					xprintf(PSTR("Erase block: %lu sectors\n"), p2);
				}
				if (disk_ioctl((BYTE)p1, MMC_GET_TYPE, &b1) == RES_OK) {
					xprintf(PSTR("Card type: %u\n"), b1);
				}
				if (disk_ioctl((BYTE)p1, MMC_GET_CSD, Buff) == RES_OK) {
					xputs(PSTR("CSD:\n")); put_dump(Buff, 0, 16);
				}
				if (disk_ioctl((BYTE)p1, MMC_GET_CID, Buff) == RES_OK) {
					xputs(PSTR("CID:\n")); put_dump(Buff, 0, 16);
				}
				if (disk_ioctl((BYTE)p1, MMC_GET_OCR, Buff) == RES_OK) {
					xputs(PSTR("OCR:\n")); put_dump(Buff, 0, 4);
				}
				if (disk_ioctl((BYTE)p1, MMC_GET_SDSTAT, Buff) == RES_OK) {
					xputs(PSTR("SD Status:\n"));
					for (s1 = 0; s1 < 64; s1 += 16) put_dump(Buff + s1, s1, 16);
				}
				if (disk_ioctl((BYTE)p1, ATA_GET_MODEL, line) == RES_OK) {
					line[40] = '\0'; xprintf(PSTR("Model: %s\n"), line);
				}
				if (disk_ioctl((BYTE)p1, ATA_GET_SN, line) == RES_OK) {
					line[20] = '\0'; xprintf(PSTR("S/N: %s\n"), line);
				}
				break;

			case 'c':	/* Disk ioctl */
				switch (*ptr++) {
				case 's':	/* dcs <pd#> - CTRL_SYNC */
					if (!xatoi(&ptr, &p1)) break;
					xprintf(PSTR("rc=%d\n"), disk_ioctl((BYTE)p1, CTRL_SYNC, 0));
					break;
				}
				break;
			}
			break;

		case 'b':
			switch (*ptr++) {
			case 'd':	/* bd <addr> - Dump R/W buffer */
				if (!xatoi(&ptr, &p1)) break;
				for (bp = &Buff[p1], ofs = p1, cnt = 32; cnt; cnt--, ptr += 16, ofs += 16) {
					put_dump(bp, ofs, 16);
				}
				break;

			case 'e':	/* be <addr> [<data>] ... - Edit R/W buffer */
				if (!xatoi(&ptr, &p1)) break;
				if (xatoi(&ptr, &p2)) {
					do {
						Buff[p1++] = (BYTE)p2;
					} while (xatoi(&ptr, &p2));
					break;
				}
				for (;;) {
					xprintf(PSTR("%04X %02X-"), (WORD)p1, Buff[p1]);
					get_line(line, sizeof line);
					ptr = line;
					if (*ptr == '.') break;
					if (*ptr < ' ') { p1++; continue; }
					if (xatoi(&ptr, &p2)) {
						Buff[p1++] = (BYTE)p2;
					}
					else {
						xputs(PSTR("???\n"));
					}
				}
				break;

			case 'r':	/* br <pd#> <sector> [<n>] - Read disk into R/W buffer */
				if (!xatoi(&ptr, &p1) || !xatoi(&ptr, &p2)) break;
				if (!xatoi(&ptr, &p3)) p3 = 1;
				xprintf(PSTR("rc=%u\n"), disk_read((BYTE)p1, Buff, p2, p3));
				break;

			case 'w':	/* bw <pd#> <sector> [<n>] - Write R/W buffer into disk */
				if (!xatoi(&ptr, &p1) || !xatoi(&ptr, &p2)) break;
				if (!xatoi(&ptr, &p3)) p3 = 1;
				xprintf(PSTR("rc=%u\n"), disk_write((BYTE)p1, Buff, p2, p3));
				break;

			case 'f':	/* bf <n> - Fill working buffer */
				if (!xatoi(&ptr, &p1)) break;
				memset(Buff, (BYTE)p1, sizeof Buff);
				break;

			case 'l':	/* bl <len> - Set read/write size for fr/fw command */
				if (!xatoi(&ptr, &p1) || (UINT)p1 == 0 || (UINT)p1 > sizeof Buff) break;
				blen = (UINT)p1;
				xprintf(PSTR("R/W length = %u\n"), blen);
				break;
			}
			break;


		case 'f':
			switch (*ptr++) {

			case 'i':	/* fi <ld#> [<mount>]- Initialize logical drive */
				if (!xatoi(&ptr, &p1) || (UINT)p1 > 9)
				{//RJA adding instructions
					printf("You must provide a drive number, such as 0.  Try 'fi 0' to mount uSD card.\n");
					break;
				}
				if (!xatoi(&ptr, &p2)) p2 = 0;
				sprintf(line, PSTR("%u:"), (UINT)p1);
				printf("Mounting drive %s with option= %d\n", line, p2);
				put_rc(f_mount(&FatFs[p1], line, (BYTE)p2));
				break;

			case 's':	/* fs [<path>] - Show logical drive status */
				while (*ptr == ' ') ptr++;
				ptr2 = ptr;
				fr = f_getfree(ptr, (DWORD*)&p2, &fs);
				if (fr) {
					put_rc(fr); break;
				}

				//RJA divided up a big printf into three parts to compile with fastspin...
				printf("FAT type = %s\nBytes/Cluster = %lu\nNumber of FATs = %u\n", fst[fs->fs_type], (DWORD)fs->csize * 512, fs->n_fats);
				printf("Root DIR entries = %u\nSectors/FAT = %lu\nNumber of clusters = %lu\n", fs->n_rootdir, fs->fsize, fs->n_fatent - 2);
				printf("Volume start (lba) = %lu\nFAT start (lba) = %lu\nDIR start (lba,clustor) = %lu\nData start (lba) = %lu\n\n", (DWORD)fs->volbase, (DWORD)fs->fatbase, (DWORD)fs->dirbase, (DWORD)fs->database);

#if FF_USE_LABEL
				fr = f_getlabel(ptr2, (char*)Buff, (DWORD*)&p1);
				if (fr) {
					put_rc(fr); break;
				}
				xprintf(Buff[0] ? PSTR("Volume name is %s\n") : PSTR("No volume label\n"), Buff);
				xprintf(PSTR("Volume S/N is %04X-%04X\n"), (WORD)((DWORD)p1 >> 16), (WORD)(p1 & 0xFFFF));
#endif
				xputs(PSTR("..."));
				AccSize = AccFiles = AccDirs = 0;
				strcpy((char*)Buff, ptr);
				fr = scan_files((char*)Buff);
				if (fr) {
					put_rc(fr); break;
				}
				xprintf(PSTR("\r%u files, %lu bytes.\n%u folders.\n"
					"%lu KB total disk space.\n%lu KB available.\n"),
					AccFiles, AccSize, AccDirs,
					(fs->n_fatent - 2) * fs->csize / 2, p2 * fs->csize / 2
				);
				break;

			case 'l':	/* fl [<path>] - Directory listing */
				while (*ptr == ' ') ptr++;
				fr = f_opendir(&Dir, ptr);
				if (fr) {
					put_rc(fr); break;
				}
				p1 = s1 = s2 = 0;
				for (;;) {
					fr = f_readdir(&Dir, &Finfo);
					if ((fr != FR_OK) || !Finfo.fname[0]) break;
					if (Finfo.fattrib & AM_DIR) {
						s2++;
					}
					else {
						s1++; p1 += Finfo.fsize;
					}
					xprintf(PSTR("%c%c%c%c%c %u/%02u/%02u %02u:%02u %9lu  %s\n"),
						(Finfo.fattrib & AM_DIR) ? 'D' : '-',
						(Finfo.fattrib & AM_RDO) ? 'R' : '-',
						(Finfo.fattrib & AM_HID) ? 'H' : '-',
						(Finfo.fattrib & AM_SYS) ? 'S' : '-',
						(Finfo.fattrib & AM_ARC) ? 'A' : '-',
						(Finfo.fdate >> 9) + 1980, (Finfo.fdate >> 5) & 15, Finfo.fdate & 31,
						(Finfo.ftime >> 11), (Finfo.ftime >> 5) & 63,
						(DWORD)Finfo.fsize,
						Finfo.fname);
				}
				if (fr == FR_OK) {
					xprintf(PSTR("%4u File(s),%10lu bytes total\n%4u Dir(s)"), s1, p1, s2);
					if (f_getfree(ptr, (DWORD*)&p1, &fs) == FR_OK) {
						xprintf(PSTR(", %10luKiB free\n"), p1 * fs->csize / 2);
					}
				}
				if (fr) put_rc(fr);
				break;

			case 'o':	/* fo <mode> <name> - Open a file */
				if (!xatoi(&ptr, &p1)) break;
				while (*ptr == ' ') ptr++;
				printf("RJA:  Opening file %s with mode %d\n", ptr, p1);
				put_rc(f_open(&File[0], ptr, (BYTE)p1));
				break;

			case 'c':	/* fc - Close a file */
				put_rc(f_close(&File[0]));
				break;

			case 'e':	/* fe - Seek file pointer */
				if (!xatoi(&ptr, &p1)) break;
				fr = f_lseek(&File[0], p1);
				put_rc(fr);
				if (fr == FR_OK) {
					xprintf(PSTR("fptr = %lu(0x%lX)\n"), (DWORD)File[0].fptr, (DWORD)File[0].fptr);
				}
				break;

			case 'r':	/* fr <len> - read file */
				if (!xatoi(&ptr, &p1)) break;
				p2 = 0;
				//RJA cli(); Timer = 0; sei();
				while (p1) {
					if (p1 >= (int) blen) { cnt = blen; p1 -= blen; }
					else { cnt = (WORD)p1; p1 = 0; }
					fr = f_read(&File[0], Buff, cnt, &s2);
					if (fr != FR_OK) { put_rc(fr); break; }
					p2 += s2;
					if (cnt != s2) break;
				}
				//RJA cli(); s2 = Timer; sei();
				xprintf(PSTR("%lu bytes read at %lu bytes/sec.\n"), p2, s2 ? (p2 * 100 / s2) : 0);
				break;

			case 'd':	/* fd <len> - read and dump file from current fp */
				if (!xatoi(&ptr, &p1)) break;
				ofs = File[0].fptr;
				while (p1) {
					if (p1 >= 16) {
						cnt = 16; p1 -= 16;
					}
					else {
						cnt = (WORD)p1; p1 = 0;
					}
					fr = f_read(&File[0], Buff, cnt, &cnt);
					if (fr != FR_OK) {
						put_rc(fr); break;
					}
					if (!cnt) break;
					put_dump(Buff, ofs, cnt);
					ofs += 16;
				}
				break;

			case 'w':	/* fw <len> <val> - write file */
				if (!xatoi(&ptr, &p1) || !xatoi(&ptr, &p2)) break;
				memset(Buff, (BYTE)p2, blen);
				p2 = 0;
				//RJA cli(); Timer = 0; sei();
				while (p1) {
					if (p1 >= (int) blen) { cnt = blen; p1 -= blen; }
					else { cnt = (WORD)p1; p1 = 0; }
					fr = f_write(&File[0], Buff, cnt, &s2);
					if (fr != FR_OK) {
						put_rc(fr); break;
					}
					p2 += s2;
					if (cnt != s2) break;
				}
				//RJA cli(); s2 = Timer; sei();
				xprintf(PSTR("%lu bytes written at %lu bytes/sec.\n"), p2, s2 ? (p2 * 100 / s2) : 0);
				break;

			case 'v':	/* fv - Truncate file */
				put_rc(f_truncate(&File[0]));
				break;

			case 'n':	/* fn <old_name> <new_name> - Change file/dir name */
				while (*ptr == ' ') ptr++;
				ptr2 = strchr(ptr, ' ');
				if (!ptr2) break;
				*ptr2++ = 0;
				while (*ptr2 == ' ') ptr2++;
				put_rc(f_rename(ptr, ptr2));
				break;

			case 'u':	/* fu <name> - Unlink a file or dir */
				while (*ptr == ' ') ptr++;
				put_rc(f_unlink(ptr));
				break;

			case 'k':	/* fk <name> - Create a directory */
				while (*ptr == ' ') ptr++;
				put_rc(f_mkdir(ptr));
				break;
#if FF_USE_EXPAND
			case 'h':	/* fh <fsz> <opt> - Allocate contiguous block */
				if (!xatoi(&ptr, &p1) || !xatoi(&ptr, &p2)) break;
				fr = f_expand(&File[0], (DWORD)p1, (BYTE)p2);
				put_rc(fr);
				break;
#endif
#if FF_USE_CHMOD
			case 'a':	/* fa <atrr> <mask> <name> - Change file/dir attribute */
				if (!xatoi(&ptr, &p1) || !xatoi(&ptr, &p2)) break;
				while (*ptr == ' ') ptr++;
				put_rc(f_chmod(ptr, p1, p2));
				break;

			case 't':	/* ft <year> <month> <day> <hour> <min> <sec> <name> */
				if (!xatoi(&ptr, &p1) || !xatoi(&ptr, &p2) || !xatoi(&ptr, &p3)) break;
				Finfo.fdate = ((p1 - 1980) << 9) | ((p2 & 15) << 5) | (p3 & 31);
				if (!xatoi(&ptr, &p1) || !xatoi(&ptr, &p2) || !xatoi(&ptr, &p3)) break;
				Finfo.ftime = ((p1 & 31) << 11) | ((p2 & 63) << 5) | ((p3 >> 1) & 31);
				while (*ptr == ' ') ptr++;
				put_rc(f_utime(ptr, &Finfo));
				break;
#endif
			case 'x': /* fx <src_name> <dst_name> - Copy file */
				while (*ptr == ' ') ptr++;
				ptr2 = strchr(ptr, ' ');
				if (!ptr2) break;
				*ptr2++ = 0;
				while (*ptr2 == ' ') ptr2++;
				xprintf(PSTR("Opening \"%s\""), ptr);
				fr = f_open(&File[0], ptr, FA_OPEN_EXISTING | FA_READ);
				if (fr) {
					put_rc(fr); break;
				}
				xprintf(PSTR("\nCreating \"%s\""), ptr2);
				fr = f_open(&File[1], ptr2, FA_CREATE_ALWAYS | FA_WRITE);
				if (fr) {
					put_rc(fr);
					f_close(&File[0]);
					break;
				}
				xprintf(PSTR("\nCopying..."));
				//RJA cli(); Timer = 0; sei();
				p1 = 0;
				for (;;) {
					fr = f_read(&File[0], Buff, sizeof Buff, &s1);
					if (fr || s1 == 0) break;   /* error or eof */
					fr = f_write(&File[1], Buff, s1, &s2);
					p1 += s2;
					if (fr || s2 < s1) break;   /* error or disk full */
				}
				if (fr) put_rc(fr);
				//RJA cli(); s2 = Timer; sei();
				xprintf(PSTR("\n%lu bytes copied at %lu bytes/sec.\n"), p1, p1 * 100 / s2);
				f_close(&File[0]);
				f_close(&File[1]);
				break;
#if FF_FS_RPATH
			case 'g':	/* fg <path> - Change current directory */
				while (*ptr == ' ') ptr++;
				put_rc(f_chdir(ptr));
				break;
#if FF_VOLUMES >= 2
			case 'j':	/* fj <path> - Change current drive */
				while (*ptr == ' ') ptr++;
				put_rc(f_chdrive(ptr));
				break;
#endif
#if FF_FS_RPATH >= 2
			case 'q':	/* fq - Show current dir path */
				fr = f_getcwd(line, sizeof line);
				if (fr) {
					put_rc(fr);
				}
				else {
					xprintf(PSTR("%s\n"), line);
				}
				break;
#endif
#endif

#if FF_USE_LABEL
			case 'b':	/* fb <name> - Set volume label */
				while (*ptr == ' ') ptr++;
				put_rc(f_setlabel(ptr));
				break;
#endif
#if FF_USE_MKFS
			case 'm':	/* fm <ld#> <type> <bytes/clust> - Create filesystem */
			{
				MKFS_PARM opt, * popt = 0;

				if (!xatoi(&ptr, &p1) || (UINT)p1 > 9) break;
				if (xatoi(&ptr, &p2)) {
					memset(&opt, 0, sizeof opt);
					popt = &opt;
					popt->fmt = (BYTE)p2;
					if (xatoi(&ptr, &p2)) {
						popt->au_size = p2;
						if (xatoi(&ptr, &p2)) {
							popt->align = p2;
							if (xatoi(&ptr, &p2)) {
								popt->n_fat = (BYTE)p2;
								if (xatoi(&ptr, &p2)) {
									popt->n_root = p2;
								}
							}
						}
					}
				}
				xprintf(PSTR("The drive %u will be formatted. Are you sure? (Y/n)="), (WORD)p1);
				get_line(line, sizeof line);
				if (line[0] == 'Y') {
					xsprintf(line, PSTR("%u:"), (UINT)p1);
					put_rc(f_mkfs(line, popt, Buff, sizeof Buff));
				}
			}
			break;
#endif


		}
		break;

#ifdef SOUND_DEFINED
		case 'p':	/* p <wavfile> - Play RIFF-WAV file */
			while (*ptr == ' ') ptr++;
			fr = f_open(&File[0], ptr, FA_READ);
			if (fr) {
				put_rc(fr);
			}
			else {
				load_wav(&File[0], "**** WAV PLAYER ****", Buff, sizeof Buff);
				f_close(&File[0]);
			}
			break;
#endif
		case 't':	/* t [<year> <mon> <mday> <hour> <min> <sec>] */
			printf("Showing time...\n");
			if (!RtcOk) break;
			if (xatoi(&ptr, &p1)) 
			{				
				rtc.year = (WORD)p1;
				xatoi(&ptr, &p1); rtc.month = (BYTE)p1;
				xatoi(&ptr, &p1); rtc.mday = (BYTE)p1;
				xatoi(&ptr, &p1); rtc.hour = (BYTE)p1;
				xatoi(&ptr, &p1); rtc.min = (BYTE)p1;
				if (!xatoi(&ptr, &p1)) break;
				rtc.sec = (BYTE)p1;
				printf("Setting Time:  %u/%u/%u %02u:%02u:%02u\n", rtc.year, rtc.month, rtc.mday, rtc.hour, rtc.min, rtc.sec);
				rtc_settime(&rtc);
			}
			rtc_gettime(&rtc);
			printf(PSTR("Year-Month-Day Hours:Minutes:Seconds = %u-%u-%u %02u:%02u:%02u\n"), rtc.year, rtc.month, rtc.mday, rtc.hour, rtc.min, rtc.sec);
			break;




		case '?':	/* Show Command List */
			xputs(PSTR(
				"[Disk contorls]\n"
				" di <pd#> - Initialize disk\n"
				" dd [<pd#> <sect>] - Dump a secrtor\n"
				" ds <pd#> - Show disk status\n"
				" dcs <pd#> - ioctl(CTRL_SYNC)\n"
				"[Buffer controls]\n"
				" bd <ofs> - Dump working buffer\n"
				" be <ofs> [<data>] ... - Edit working buffer\n"
				" br <pd#> <sect> [<num>] - Read disk into working buffer\n"
				" bw <pd#> <sect> [<num>] - Write working buffer into disk\n"
				" bf <val> - Fill working buffer\n"
				" bl <len> - Set read/write length for fr/fw command\n"
				"[Filesystem controls]\n"
				" fi <ld#> [<mount>] - Force initialized the volume\n"
				" fs [<path>] - Show volume status\n"
				" fl [<path>] - Show a directory\n"
				" fo <mode> <file> - Open a file\n"
				" fc - Close the file\n"
				" fe <ofs> - Move fp in normal seek\n"
				" fh <fsz> <opt> - Allocate a contiguous block to the file\n"
				" fd <len> - Read and dump the file\n"
				" fr <len> - Read the file\n"
				" fw <len> <val> - Write to the file\n"
				" fn <org name> <new name> - Rename an object\n"
				" fu <obj name> - Unlink an object\n"
				" fv - Truncate the file at current fp\n"
				" fk <dir name> - Create a directory\n"
				" fa <atrr> <mask> <object name> - Change object attribute\n"
				" ft <year> <month> <day> <hour> <min> <sec> <object name> - Change timestamp of an object\n"
				" fx <src file> <dst file> - Copy a file\n"
				" fg <path> - Change current directory\n"
				" fj <path> - Change current drive\n"
				" fq - Show current directory\n"
				" fm <ld#> [<fs type> [<au size> [<align> [<N fats> [<N root>]]]]] - Create FAT volume\n"
				"[Misc commands]\n"
				" p <wavfile> - Play RIFF-WAVE file\n"
				" t [<year> <month> <mday> <hour> <min> <sec>] - Set/Show current time\n"
				"\n"));
			break;

		}
	}
}
