//RJA modifying this code for Propeller II
/*------------------------------------------------------------------------/
/  Sound Streamer and RIFF-WAVE file player for Mega AVRs
/-------------------------------------------------------------------------/
/
/  Copyright (C) 2011, ChaN, all right reserved.
/
/ * This software is a free software and there is NO WARRANTY.
/ * No restriction on use. You can use, modify and redistribute it for
/   personal, non-profit or commercial products UNDER YOUR RESPONSIBILITY.
/ * Redistributions of source code must retain the above copyright notice.
/
/-------------------------------------------------------------------------*/
/* Note that the audio playback function is a heavy load for 8-bit MCUs.
/  High bitrate file in CD quality will be played as buzz sound due to
/  buffer underrun. If it is the case, the sound file should be re-sampled
/  to lower sampling frequency, 8-bit.
*/ //Note:  P2 has no problem with CD quality files :)


#include <string.h>
//#include <avr/io.h>
//#include <avr/interrupt.h>
#include "sound.h"
//#include "xitoa.h"
//#include "uart.h"
#include "propeller2.h"
#include "PropPins.h"
#include <stdio.h>

// the Spin2 wav player
struct __using("WavPlayer.spin2") WavPlayer; //RJA:  Visual Studio will show an error here, but it's OK


//RJA fixing some things...
#define PSTR  //RJA replacing this with nothing...
#define xputc putchar
#define xgetc getchar
//#define xatoi atoi //Actually need to define a function for this...
#define xprintf printf
#define xputs puts


#define NBSIZE 32

#define FCC(c1,c2,c3,c4)	(((DWORD)c4<<24)+((DWORD)c3<<16)+(c2<<8)+c1)	/* FourCC */
#define	LD_WORD(ptr)		(WORD)(*(WORD*)(BYTE*)(ptr))
#define	LD_DWORD(ptr)		(DWORD)(*(DWORD*)(BYTE*)(ptr))



static
WAVFIFO *WavFifo;	/* Pointer to sound FIFO control block */




/*-----------------------------------------------------*/
/* WAV file loader                                     */

int load_wav (
	FIL *fp,			/* Pointer to the open file object to play */
	const char *title,	/* Title (file name, etc...) */
	void *work,			/* Pointer to working buffer (must be-4 byte aligned) */
	UINT sz_work		/* Size of working buffer (must be power of 2) */
)
{

	UINT md, wi, br, tc, t, btr;
	DWORD sz, ssz, offw, szwav, wsmp, fsmp, eof;
	WAVFIFO fcb;
	BYTE k, *buff = work;
	char* p, nam[NBSIZE], art[NBSIZE];


	xprintf(PSTR("%s\n"), title);	/* Put title */

	/* Is it a WAV file? */
	if (f_read(fp, buff, 12, &br) || br != 12) return -1;
	if (LD_DWORD(&buff[0]) != FCC('R','I','F','F')) return -1;
	if (LD_DWORD(&buff[8]) != FCC('W','A','V','E')) return -1;
	eof = LD_DWORD(&buff[4]) + 8;

	/* Analyze the RIFF-WAVE header and get properties */
	nam[0] = art[0] = 0;
	md = fsmp = wsmp = offw = szwav = 0;
	while (f_tell(fp) < eof) {
		if (f_read(fp, buff, 8, &br) || br != 8) return -1;
		sz = (LD_DWORD(&buff[4]) + 1) & ~1;
		switch (LD_DWORD(&buff[0])) {
		case FCC('f','m','t',' ') :
			if (sz > 1000 || sz < 16 || f_read(fp, buff, sz, &br) || sz != br) return -1;
			if (LD_WORD(&buff[0]) != 0x1) return -1;	/* Check if LPCM */
			if (LD_WORD(&buff[2]) == 2) {	/* Channels (1 or 2) */
				md = 1; wsmp = 2;
			} else {
				md = 0; wsmp = 1;
			}
			if (LD_WORD(&buff[14]) == 16) {	/* Resolution (8 or 16) */
				md |= 2; wsmp *= 2;
			}
			fsmp = LD_DWORD(&buff[4]);		/* Sampling rate */
			break;

		case FCC('f','a','c','t') :
			f_lseek(fp, f_tell(fp) + sz);
			break;

		case FCC('d','a','t','a') :
			offw = f_tell(fp);	/* Wave data start offset */
			szwav = sz;			/* Wave data length [byte] */
			f_lseek(fp, f_tell(fp) + sz);
			break;

		case FCC('L','I','S','T'):
			sz += f_tell(fp);
			if (f_read(fp, buff, 4, &br) || br != 4) return -1;
			if (LD_DWORD(buff) == FCC('I','N','F','O')) {	/* LIST/INFO chunk */
				while (f_tell(fp) < sz) {
					if (f_read(fp, buff, 8, &br) || br != 8) return -1;
					ssz = (LD_DWORD(&buff[4]) + 1) & ~1;
					p = 0;
					switch (LD_DWORD(buff)) {
					case FCC('I','N','A','M'):		/* INAM sub-chunk */
						p = nam; break;
					case FCC('I','A','R','T'):		/* IART sub-cnunk */
						p = art; break;
					}
					if (p && ssz <= NBSIZE) {
						if (f_read(fp, p, ssz, &br) || br != ssz) return -1;
					} else {
						if (f_lseek(fp, f_tell(fp) + ssz)) return -1;
					}
				}
			} else {
				if (f_lseek(fp, sz)) return -1;	/* Skip unknown sub-chunk type */
			}
			break;

		default :	/* Unknown chunk */
			return -1;
		}
	}
	if (!szwav || !fsmp) return -1;		/* Check if valid WAV file */
	if (f_lseek(fp, offw)) return -1;	/* Seek to top of wav data */
	tc = (UINT)(szwav / fsmp / wsmp);	/* Length (sec) */

	xprintf(PSTR("IART=%s\nINAM=%s\n"), art, nam);
	xprintf(PSTR("Sample=%u.%ukHz/%ubit/%S\nLength=%u:%02u\n"), (UINT)(fsmp / 1000), (UINT)(fsmp / 100) % 10, (md & 2) ? 16 : 8, (md & 1) ? PSTR("st") : PSTR("mo"), tc / 60, tc % 60);

	/* Initialize stream parameters and start sound streming */
	fcb.mode = md;
	fcb.buff = buff;
	fcb.sz_buff = sz_work;
	//RJA if (!sound_start(&fcb, fsmp)) return -1;

	//Fill the two buffers before starting the player
	int BuffSize = WavPlayer.GetBufferSize(); //Does not include the empty flags at end of buffer
	uint16_t *b1 = WavPlayer.GetBuffer1();
	uint16_t *b2 = WavPlayer.GetBuffer2();
	f_read(fp, b1, BuffSize  *2, &br);
	//printf("%d bytes read into buffer 1\n", br);
	f_read(fp, b2, BuffSize * 2, &br);
	//printf("%d bytes read into buffer 2\n", br);
	//clear buffer empty flags (the last word in the buffer)
	b1[BuffSize] = 0;
	b2[BuffSize] = 0;
	

	//Start the assembly wave player
	int nSamples = (szwav / 2) / (wsmp/2);
	printf("Starting wav player with Sample Rate=%d, #Samples=%d, #Channels=%d\n", fsmp, nSamples, wsmp/2);
	WavPlayer.SetPins(AV_LeftPin, AV_RightPin);
	WavPlayer.StartWavPlayerCog(fsmp, nSamples, (wsmp/2));//(SampleRate, Samples, nChannels)
	_waitms(100);
	printf("Playing...\n");
	WavPlayer.GiveAtn();

	int ntr = BuffSize * 2;

	while (1)
	{
		
		while (b1[BuffSize] != 1) {}//Note:  There is actually one extra word at end of buffer that is used as a flag
		b1[BuffSize] = 0;
		f_read(fp, b1, ntr, &br);	
		if (br != ntr)
			break; //end of file?
		WavPlayer.GiveAtn();
		while (b2[BuffSize] != 1) {}
		b2[BuffSize] = 0;
		f_read(fp, b2, ntr, &br);
		if (br != ntr)  //end of file?
			break;
	}

	printf("Playback complete.\n");


	return 0;	/* Terminated due to -1:error, 0:eot, >0:key code */
}


